#include <iostream>
#include <iomanip>
#include <algorithm>
#include "MeshLoader.h"
#include "Predicates.h"

MeshLoader::MeshLoader() : fileName("") {}

MeshLoader::MeshLoader(const std::string& p_fileName) : fileName(p_fileName) {}

MeshLoader::~MeshLoader()
{
    //!!! Для полей и так деструктор вызовется
    
	nodes.clear();
	finiteElements.clear();
	boundaryFiniteElements.clear();
	fileName.clear();
}

void MeshLoader::printNodes()
{
	
	std::for_each(nodes.begin(), nodes.end(), [](const Node& node) {
		std::cout << "ID: " << std::setw(4) << node.ID << "| ";
		std::cout << std::fixed << std::setprecision(6);

		std::for_each(node.coords.begin(), node.coords.end(), [](const auto& coord) {
			std::cout << std::setw(10) << coord;
			});
		std::cout << std::endl; 
	});
}

void MeshLoader::printFiniteElems()
{
	std::for_each(finiteElements.begin(), finiteElements.end(), [](const auto& fe) {
		std::cout << "ID: " << std::setw(4) << fe.second.ID << "| Material ID: " << std::setw(3)
			<< fe.second.materialId << "| " << "Nodes ID's: ";
		std::for_each(fe.second.nodes.begin(), fe.second.nodes.end(), [](const auto& id) {
			std::cout << std::setw(4) << id;
			});
		std::cout << std::endl;
	});
}

void MeshLoader::printBoundaryFiniteElements()
{
	std::for_each(boundaryFiniteElements.begin(), boundaryFiniteElements.end(), [](const auto& bfe) {
		std::cout << "ID: " << std::setw(4) << bfe.second.ID << "| Surface ID: " << std::setw(3)
			<< bfe.second.surfaceId << "| " << "Nodes ID's: ";

		std::for_each(bfe.second.nodes.begin(), bfe.second.nodes.end(), [](auto& nodeIds) {
			std::cout << std::setw(4) << nodeIds;
		});
		std::cout << std::endl;
		}
	);
}

const std::vector<Node>& MeshLoader::getNodes()
{
	return nodes;
}

const std::unordered_map<uint16_t, FiniteElement>& MeshLoader::getFiniteElems()
{
	return finiteElements;
}

const std::unordered_map<uint16_t, BoundaryFiniteElement>& MeshLoader::getBoundaryFiniteElements()
{ 
	return boundaryFiniteElements;
}

std::vector<FiniteElement> MeshLoader::getFEbyNodesIds(uint16_t id1, uint16_t id2, uint16_t id3)
{
	std::vector<FiniteElement> answer;
	auto currentIter = finiteElements.begin();
	auto endIter = finiteElements.end();
	while (currentIter != endIter) {
		findFiniteElementBy3Nodes finder(id1,id2,id3);
		currentIter = std::find_if(currentIter, endIter, finder);
		if (currentIter != endIter) {
			answer.push_back((*currentIter).second);
			++currentIter;
		}
	}
	answer.shrink_to_fit();
	return answer;
}

std::vector<std::unordered_map<uint16_t, FiniteElement>::iterator> MeshLoader::getFEbyEdge(uint16_t id1, uint16_t id2)
{
	std::vector<std::unordered_map<uint16_t, FiniteElement>::iterator> answer;
	auto currentIter = finiteElements.begin();
	auto endIter = finiteElements.end();
	while (currentIter != endIter) {
		findFiniteElementByEdge finder(id1, id2);
		currentIter = std::find_if(currentIter, endIter, finder);
		if (currentIter != endIter) {
			answer.push_back(currentIter);
			++currentIter;
		}
	}
	answer.shrink_to_fit();
	return answer;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBFEbySurfaceID(uint16_t surID)
{
	std::vector<BoundaryFiniteElement> answer;
	auto currentIter = boundaryFiniteElements.begin();
	auto endIter = boundaryFiniteElements.end();
	while (currentIter != endIter) {
		findBoundaryBySurfaceID finder(surID);
		currentIter = std::find_if(currentIter, endIter, finder);
		if (currentIter != endIter) {
			answer.push_back((*currentIter).second);
			++currentIter;
		}
	}
	answer.shrink_to_fit();
	return answer;
}

std::vector<std::unordered_map<uint16_t, BoundaryFiniteElement>::iterator> MeshLoader::getBFEbyEdge(uint16_t id1, uint16_t id2)
{
	std::vector<std::unordered_map<uint16_t, BoundaryFiniteElement>::iterator> answer;
	auto currentIter = boundaryFiniteElements.begin();
	auto endIter = boundaryFiniteElements.end();
	while (currentIter != endIter) {
		findBoundaryFiniteElementByEdge finder(id1, id2);
		currentIter = std::find_if(currentIter, endIter, finder);
		if (currentIter != endIter) {
			answer.push_back(currentIter);
			++currentIter;
		}
	}
	answer.shrink_to_fit();
	return answer;
}

std::vector<Node> MeshLoader::getNodesbySurfaceID (uint16_t surID)
{
	std::vector<Node> answer;
	std::vector<BoundaryFiniteElement> temp = this->getBFEbySurfaceID(surID);
	auto currentIter = temp.begin();
	auto endIter = temp.end();
	while (currentIter != endIter) {
		answer.push_back(nodes.at((currentIter->nodes.at(0)) - 1));
		answer.push_back(nodes.at((currentIter->nodes.at(1)) - 1));
		answer.push_back(nodes.at((currentIter->nodes.at(2)) - 1));
		++currentIter;
	}
	std::sort(answer.begin(), answer.end(), [](const Node& p_node_1, const Node& p_node_2) { return p_node_1.ID < p_node_2.ID; });
	answer.erase(std::unique(answer.begin(), answer.end(), [](const Node& p_node_1, const Node& p_node_2) { return p_node_1.ID == p_node_2.ID; }), answer.end());
	answer.shrink_to_fit();
	return answer;
}

std::vector<FiniteElement> MeshLoader::getFEbyMaterialID(uint16_t areaID)
{
	std::vector<FiniteElement> answer;
	auto currentIter = finiteElements.begin();
	auto endIter = finiteElements.end();
	while (currentIter != endIter)
	{
		findFiniteElementByMaterialID finder(areaID);
		currentIter = std::find_if(currentIter, endIter, finder);
		if (currentIter != endIter) {
			answer.push_back((*currentIter).second);
			++currentIter;
		}
	}
	answer.shrink_to_fit();
	return answer;

}

std::vector<std::set<uint16_t>> MeshLoader::nodeNeighbours()
{
	std::vector<std::set<uint16_t>> result(nodesCount + 1);
	std::for_each(this->finiteElements.begin(), this->finiteElements.end(), [&result](std::pair<const uint16_t, FiniteElement>& fe)
		{
			std::for_each(fe.second.nodes.begin(), fe.second.nodes.end(), [&fe, &result](uint16_t currID)
				{
					std::copy_if(fe.second.nodes.begin(), fe.second.nodes.end(), std::inserter(result.at(currID), result.at(currID).begin()), [&currID](uint16_t thisID)
						{
							return thisID != currID;
						});
				});
		});
	result.erase(result.begin());
    return result;
}

bool MeshLoader::insertMiddlesOfEdges()
{
	uint16_t id1;
	uint16_t id2;
	auto currentIter = finiteElements.begin();
	auto endIter = finiteElements.end();
	while (currentIter != endIter) {
		for (int i = 0; i < 4; i++) {
			for (int j = i + 1; j < 4; j++) {

				id1 = currentIter->second.nodes.at(i);
				id2 = currentIter->second.nodes.at(j);

				if ((alreadyHasMiddle.find(std::make_pair(id1, id2)) == alreadyHasMiddle.end())
					|| (alreadyHasMiddle.find(std::make_pair(id2, id1)) == alreadyHasMiddle.end())) {
					//!!! Чтобы не делать две проверки см. замечание в хешь-функции.
					    
					Node newNode;
					newNode.ID = nodesCount + 1;
					for (unsigned int k = 0; k < 3; k++)
					{
						newNode.coords.push_back((this->nodes.at(id1).coords.at(i) + this->nodes.at(id2).coords.at(k)) / 2);
					}
					nodes.push_back(newNode);
					nodesCount++;

					if (std::find(boundaryEdges.begin(), boundaryEdges.end(), std::make_pair(id1, id2)) != boundaryEdges.end() ||
						std::find(boundaryEdges.begin(), boundaryEdges.end(), std::make_pair(id2, id1)) != boundaryEdges.end()) {
						auto bfe = getBFEbyEdge(id1, id2);
						for (int i = 0; i < bfe.size(); i++) {
							bfe.at(i)->second.nodes.push_back(newNode.ID);
						}
					}

					auto fe = getFEbyEdge(id1, id2);
					for (int i = 0; i < fe.size(); i++) {
						fe.at(i)->second.nodes.push_back(newNode.ID);
					}

					alreadyHasMiddle.insert(std::make_pair(id1, id2));
					alreadyHasMiddle.insert(std::make_pair(id2, id1));
				}
			}
		}
		currentIter++;
	}
	std::cout << "Nodes have been succesfully inserted!" << std::endl;
	return true;
}
